
typedef struct {

  unsigned int discrete : 1;  /* this column is discrete data. */
  unsigned int gaussian : 1;  /* this column is Gaussian data. */
  unsigned int complete : 1;  /* this column contains no missing data points. */
  unsigned int fixed    : 1;  /* this column should never be (re)moved. */
  unsigned int drop     : 1;  /* this column is to be removed. */
  unsigned int padding  : 3;  /* pad to 1 byte. */

} flags;

typedef struct {

  int nobs;             /* number of observations. */
  int ncols;            /* number of columns. */
  const char **names;   /* column names (optional). */
  flags *flag;

} meta;

void meta_init_flags(meta *m, int offset, SEXP complete, SEXP fixed);
void meta_copy_names(meta *m, int offset, SEXP df);

typedef struct {

  meta m;               /* metadata. */
  int **col;            /* pointers to the discrete columns. */
  int *nlvl;            /* number of levels of the discrete columns. */

} ddata;

ddata ddata_from_SEXP(SEXP df, int offset);
ddata empty_ddata(int nobs, int ncols);
void ddata_drop_flagged(ddata *dt, ddata *copy);
void FreeDDT(ddata dt, int free_data);

typedef struct {

  meta m;               /* metadata. */
  double **col;         /* pointers to the continuous columns. */
  double *mean;         /* means of the continuous columns (optional). */

} gdata;

gdata gdata_from_SEXP(SEXP df, int offset);
gdata empty_gdata(int nobs, int ncols);
void gdata_cache_means(gdata *dt, int offset);
void gdata_drop_column(gdata *dt, gdata *copy, int drop);
void gdata_move_column(gdata *dt, gdata *copy, int i, int j);
void FreeGDT(gdata dt, int free_data);

typedef struct {

  meta m;               /* metadata. */
  int **dcol;           /* pointers to the discrete columns. */
  double **gcol;        /* pointers to the continuous columns. */
  int *nlvl;            /* number of levels of the discrete columns. */
  int ndcols;           /* number of discrete columns. */
  int ngcols;           /* number of continuous columns. */
  int *map;             /* mapping between the original column position and the
                         * positions of the columns in the data structure. */

} cgdata;

cgdata cgdata_from_SEXP(SEXP df, int doffset, int goffset);
cgdata empty_cgdata(int nobs, int dcols, int gcols);
void print_cgdata(cgdata dt);
void cgdata_drop_flagged(cgdata *dt, cgdata *copy);
void FreeCGDT(cgdata, int free_data);
